# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: beam_grpc.proto
# Protobuf Python Version: 4.25.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0f\x62\x65\x61m_grpc.proto\x12\x04\x62\x65\x61m\"A\n\x12SetVariableRequest\x12\x0e\n\x06\x63lient\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\x0c\"&\n\x13SetVariableResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\"2\n\x12GetVariableRequest\x12\x0e\n\x06\x63lient\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\"$\n\x13GetVariableResponse\x12\r\n\x05value\x18\x01 \x01(\x0c\"U\n\x15QueryAlgorithmRequest\x12\x0e\n\x06\x63lient\x18\x01 \x01(\t\x12\x0e\n\x06method\x18\x02 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x03 \x01(\x0c\x12\x0e\n\x06kwargs\x18\x04 \x01(\x0c\")\n\x16QueryAlgorithmResponse\x12\x0f\n\x07results\x18\x01 \x01(\x0c\"\x10\n\x0eGetInfoRequest\"\x1f\n\x0fGetInfoResponse\x12\x0c\n\x04info\x18\x01 \x01(\t2\xa2\x02\n\x0b\x42\x65\x61mService\x12\x44\n\x0bSetVariable\x12\x18.beam.SetVariableRequest\x1a\x19.beam.SetVariableResponse\"\x00\x12\x44\n\x0bGetVariable\x12\x18.beam.GetVariableRequest\x1a\x19.beam.GetVariableResponse\"\x00\x12M\n\x0eQueryAlgorithm\x12\x1b.beam.QueryAlgorithmRequest\x1a\x1c.beam.QueryAlgorithmResponse\"\x00\x12\x38\n\x07GetInfo\x12\x14.beam.GetInfoRequest\x1a\x15.beam.GetInfoResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'beam_grpc_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_SETVARIABLEREQUEST']._serialized_start=25
  _globals['_SETVARIABLEREQUEST']._serialized_end=90
  _globals['_SETVARIABLERESPONSE']._serialized_start=92
  _globals['_SETVARIABLERESPONSE']._serialized_end=130
  _globals['_GETVARIABLEREQUEST']._serialized_start=132
  _globals['_GETVARIABLEREQUEST']._serialized_end=182
  _globals['_GETVARIABLERESPONSE']._serialized_start=184
  _globals['_GETVARIABLERESPONSE']._serialized_end=220
  _globals['_QUERYALGORITHMREQUEST']._serialized_start=222
  _globals['_QUERYALGORITHMREQUEST']._serialized_end=307
  _globals['_QUERYALGORITHMRESPONSE']._serialized_start=309
  _globals['_QUERYALGORITHMRESPONSE']._serialized_end=350
  _globals['_GETINFOREQUEST']._serialized_start=352
  _globals['_GETINFOREQUEST']._serialized_end=368
  _globals['_GETINFORESPONSE']._serialized_start=370
  _globals['_GETINFORESPONSE']._serialized_end=401
  _globals['_BEAMSERVICE']._serialized_start=404
  _globals['_BEAMSERVICE']._serialized_end=694
# @@protoc_insertion_point(module_scope)
